from functools import lru_cache

from .str_algo import is_ascii_alpha, is_digit, ascii_punctuations


def token_stream(file_or_files, with_postag=False, skip_space=False, use_lib='jieba'):
    if isinstance(file_or_files, str):
        files = (file_or_files,)
    else:
        files = tuple(file_or_files)

    for file in files:
        with open(file, 'r') as f:
            for line in f:
                yield from tokenizer(line,
                                     with_postag=with_postag,
                                     skip_space=skip_space,
                                     use_lib=use_lib)


@lru_cache(maxsize=32)
def _load_spacy_lang(lang):
    import spacy
    return spacy.load(lang)


def _cut(text, use_lib, lang):
    if use_lib == 'hanlp':
        from pyhanlp import HanLP
        for term in HanLP.segment(text):
            yield term.word, term.nature
    elif use_lib == 'jieba':
        import jieba.posseg as posseg
        for token in posseg.cut(text):
            yield token.word, token.flag
    elif use_lib == 'spacy':
        nlp = _load_spacy_lang(lang)
        for token in nlp(text):
            yield token.text, token.pos_
    elif use_lib == 'naive':
        acc = []
        for ch in text:
            if ch == ' ':
                if acc:
                    yield ''.join(acc), None
                acc = []
            elif ch in ascii_punctuations:
                if acc:
                    yield ''.join(acc), None
                yield ch, None
                acc = []
            else:
                acc.append(ch)
        if acc:
            yield ''.join(acc), None
    else:
        raise ValueError('only support jieba or spacy, but found:{}'.format(use_lib))


def tokenizer(text,
              with_postag=False,
              to_upper=True,
              skip_space=False,
              cut_digits=False,
              cut_ascii=False,
              use_lib='jieba',
              lang='en'):
    for word, postag in _cut(text, use_lib, lang):
        if skip_space and word == ' ':
            continue
        if to_upper:
            word = word.upper()

        if (cut_digits and all(is_digit(c) for c in word)) or (cut_ascii and all(is_ascii_alpha(c) for c in word)):
            for c in word:
                if with_postag:
                    yield c, postag
                else:
                    yield c
        else:
            if with_postag:
                yield word, postag
            else:
                yield word



import fasttext
from . import wordembedding


class WordEmbedding(wordembedding.WordEmbedding):
    def __init__(self, model_path):
        super(WordEmbedding, self).__init__()
        self._model = None
        self.model_path = model_path

    def _ensure_model_loaded(self):
        if self._model is None:
            self.model = fasttext.load_model(self.model_path)

    def __repr__(self):
        return '<fasttext({})>'.format(self.model_path)

    def get_raw_word2vec(self, word):
        self._ensure_model_loaded()
        return self._model[word]



import collections
from multiprocessing import Pool


class Vocab:

    def __init__(self, words=(), from_freqs=None, min_count=1):
        self._freqs = collections.Counter(words)
        if from_freqs is not None:
            for k, v in from_freqs.items():
                self._freqs[k] += v
        self.shrink_(min_count)

    def increase_freq(self, word, freq=1):
        self._freqs[word] += freq
        return self

    def shrink_(self, min_count):
        assert min_count >= 0
        freqs = dict((w, f) for w, f in self._freqs.items() if f >= min_count)
        self._freqs = collections.Counter(freqs)
        return self

    def merge_(self, other):
        for k, v in other.items():
            self._freqs[k] += v

    def shrink(self, min_count):
        assert min_count >= 0
        vocab = Vocab()
        for w, f in self._freqs.items():
            if f >= min_count:
                vocab[w] = f
        return vocab

    def __contains__(self, item):
        return item in self._freqs

    def __delitem__(self, key):
        del self._freqs[key]

    def __setitem__(self, key, value):
        self._freqs[key] = value

    def __getitem__(self, item):
        return self._freqs[item]

    def __iter__(self):
        return iter(self._freqs.keys())

    def __len__(self):
        return len(self._freqs)

    def __repr__(self):
        return '<Vocab(n_word={})>'.format(len(self))

    @classmethod
    def load(cls, file):
        vocab = Vocab()
        with open(file, 'r') as f:
            for line in f:
                try:
                    line = line.strip(' \n\t')
                    word, *count = line.split(' ')
                    if not word:
                        continue
                    if not count:
                        vocab.increase_freq(word)
                    else:
                        vocab.increase_freq(word, int(count[0]))
                except ValueError as e:
                    print(e)
        return vocab

    def dump(self, file):
        with open(file, 'w') as f:
            for word, freq in self._freqs.items():
                f.write('{} {}\n'.format(word, freq))

    def vocab_len(self, min_count=1):
        return len(self.shrink(min_count))

    def words(self, min_count=1):
        return tuple(self.shrink(min_count))

    def items(self):
        return self._freqs.items()


def _build_vocab_from_corpus(corpus_seq):
    from yxt_nlp_toolkit.utils.tokenizer import token_stream
    return Vocab(words=token_stream(corpus_seq))


def _batch_item(seq, batch_size):
    batch = []
    for e in seq:
        batch.append(e)
        if len(batch) >= batch_size:
            yield batch
            batch = []
    if batch:
        yield batch


def build_vocab_from_corpus(corpus_or_corpus_seq, min_count=1):
    if isinstance(corpus_or_corpus_seq, str):
        corpus_or_corpus_seq = [corpus_or_corpus_seq]

    batch_seq = list(_batch_item(corpus_or_corpus_seq, 100))
    vocab = Vocab()
    with Pool(10) as p:
        out = p.map(_build_vocab_from_corpus, batch_seq)
        for e in out:
            vocab.merge_(e)
    return vocab.shrink_(min_count=min_count)



import numpy as np


class WordEmbedding:
    def __init__(self):
        self._embedding_dim = None

    @property
    def embedding_dim(self):
        if self._embedding_dim is not None:
            return self._embedding_dim
        self._ensure_model_loaded()
        for word in ('中国', '知识', '能力'):
            try:
                self._embedding_dim = len(self.get_raw_word2vec(word))
                return self._embedding_dim
            except KeyError:
                pass
        raise ValueError('Fail to detect wordvec len')

    def __getitem__(self, item):
        self._ensure_model_loaded()
        vec = self.get_raw_word2vec(item)
        return np.array(vec)

    def _ensure_model_loaded(self):
        raise NotImplementedError

    def get_raw_word2vec(self, word):
        raise NotImplementedError



from . import wordembedding


class WordEmbedding(wordembedding.WordEmbedding):
    def __init__(self, model_path):
        super(WordEmbedding, self).__init__()
        self.model_path = model_path
        self._model = None

    def _ensure_model_loaded(self):
        import gensim
        if self._model is None:
            self._model = gensim.models.Word2Vec.load(self.model_path)

    def __repr__(self):
        return "<gensim({})>".format(self.model_path)

    def get_raw_word2vec(self, word):
        return self._model.wv[word]



import numpy as np
import pickle


class Lang:
    DEFAULT_NIL_TOKEN, NIL_INDEX = '<NIL>', 0

    def __init__(self, words, to_upper=True,
                 reserved_tokens=(), nil_token=DEFAULT_NIL_TOKEN):
        self._word2ix, self._ix2word = {}, {}
        self.to_upper = to_upper
        self._add_new_word(nil_token)
        for token in reserved_tokens:
            self._add_new_word(token)
        for word in words:
            self._add_new_word(word)

    def word_iter(self):
        return iter(self._word2ix.keys())

    def index_iter(self):
        return iter(self._ix2word.keys())

    def __len__(self):
        return len(self._word2ix)

    def __contains__(self, item):
        item = item.upper() if self.to_upper else item
        if isinstance(item, int):
            return item in self._ix2word
        elif isinstance(item, str):
            return item in self._word2ix
        else:
            return False

    def __iter__(self):
        return iter(self._word2ix.keys())

    def items(self):
        return self._word2ix.items()

    def __repr__(self):
        return str(self)

    def __str__(self):
        return 'Lang(vocab_size={vocab_size})'.format(vocab_size=self.vocab_size)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.word(item)
        elif isinstance(item, str):
            return self.ix(item)
        raise TypeError("only support int,str:but found:{}({})".format(
            type(item), item))

    def _add_new_word(self, word, index=None):
        word = word.upper() if self.to_upper else word
        if word in self._word2ix:
            if index is not None:
                assert self.ix(word) == index
            return
        index = len(self._word2ix) if index is None else index
        assert index not in self._ix2word
        self._word2ix[word], self._ix2word[index] = index, word

    def build_embedding(self, wv, out_embedding=None):
        from ..embedding.wordembedding import WordEmbedding
        if not isinstance(wv, WordEmbedding):
            raise TypeError('only support WordEmbedding,but found {}'.format(type(wv)))
        if out_embedding is None:
            out_embedding = np.random.randn(self.vocab_size, wv.embedding_dim)
        for ix, word in self._ix2word.items():
            try:
                if ix < len(out_embedding):
                    out_embedding[ix] = wv[word]
            except KeyError:
                pass
        return out_embedding

    @property
    def vocab_size(self):
        return len(self._word2ix)

    @property
    def nil_token(self):
        return self.word(0)

    def ix(self, word):
        assert isinstance(word, str)
        word = word.upper() if self.to_upper else word
        return self._word2ix.get(word, Lang.NIL_INDEX)

    def to_indices(self, words):
        return tuple(self.ix(w) for w in words)

    def to_words(self, indices):
        return tuple(self.word(i) for i in indices)

    def one_hot_of(self, word_or_index):
        vocab_len = self.vocab_size
        if isinstance(word_or_index, str):
            ix = self.ix(word_or_index)
        elif isinstance(word_or_index, int):
            ix = word_or_index
        else:
            raise TypeError("one hot only support str or int, but found:{}({})".format(
                type(word_or_index), word_or_index))

        assert 0 <= ix < vocab_len
        vec = [0] * vocab_len
        vec[ix] = 1
        return vec

    def word(self, index):
        assert isinstance(index, int)
        if index == Lang.NIL_INDEX:
            return Lang.NIL_TOKEN
        if index in self._ix2word:
            return self._ix2word[index]
        raise ValueError('unknown index:{}'.format(index))

    def vocabulary(self):
        return tuple(self._word2ix.keys())

    def dump(self, path, binary=False):
        if binary:
            with open(path, 'wb') as f:
                pickle.dump(self, f)
        else:
            with open(path, 'w') as f:
                for word, index in self._word2ix.items():
                    word = word.strip('\t ')
                    if word == '\n':
                        word = '<new_line>'
                    elif word == '\t':
                        word = '<tab>'
                    entry = '{} {}\n'.format(word, index)
                    f.write(entry)

    @classmethod
    def load(cls, path, binary=False):
        if binary:
            with open(path, 'rb') as f:
                return pickle.load(f)
        else:
            # TODO: loss the name, fix it
            lang = Lang(words=())
            lang._ix2word, lang._word2ix = {}, {}
            with open(path, 'r') as f:
                for line in f:
                    word, index = line.strip().split(' ')
                    if word == '<new_line>':
                        word = '\n'
                    elif word == '<tab>':
                        word = '\t'
                    index = int(index)
                    lang._add_new_word(word, index)
            return lang


def build_lang_from_token_stream(token_stream, min_count=1, lang_name='zh'):
    from collections import Counter
    words_freq = Counter(token_stream)
    words = tuple(w for w, freq in words_freq.items() if freq >= min_count)
    return Lang(name=lang_name, words=words)


def build_lang_from_corpus(corpus_or_corpus_seq, min_count=1, lang_name='zh'):
    from yxt_nlp_toolkit.utils.tokenizer import token_stream
    tokens = token_stream(corpus_or_corpus_seq)
    return build_lang_from_token_stream(tokens, min_count=min_count, lang_name=lang_name)



import os
import re

_current_dir = os.path.dirname(os.path.abspath(__file__))
_user_custom_dict = _current_dir + '/dict.dat'
_stopwords_data = _current_dir + '/stopwords.dat'

_loaded_dict = False


def jieba_load_userdict():
    global _loaded_dict
    import jieba
    if _loaded_dict:
        return
    jieba.load_userdict(_user_custom_dict)
    _loaded_dict = True


def read_user_custom_words():
    with open(_user_custom_dict, 'r') as f:
        for line in f:
            try:
                word, *_ = re.split('\s', line)
                yield word
            except ValueError as e:
                print(e)


def _read_stopwords():
    with open(_stopwords_data, 'r') as f:
        for line in f:
            line = line.replace('\n', '').strip()
            if line:
                yield line


_stopword_set = set(_read_stopwords())


def is_stopword(word):
    return word in _stopword_set


def get_stopwords():
    return _stopword_set



from . import wordembedding
import numpy


class WordEmbedding(wordembedding.WordEmbedding):

    def __init__(self, model_path):
        super(WordEmbedding, self).__init__()
        self.model_path = model_path
        self._model = None

    def _ensure_model_loaded(self):
        if self._model is None:
            self._model = dict(self._load(self.model_path))

    @staticmethod
    def _load(path):
        with open(path, 'r') as f:
            for line in f:
                word, *vec = line.split(' ')
                vec = numpy.array(tuple(float(v) for v in vec))
                yield word.upper(), vec

    def __repr__(self):
        return "<general({})>".format(self.model_path)

    def get_raw_word2vec(self, word):
        return self._model[word]



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='yxt_nlp_toolkit',
    version='0.2.0',
    description='common utils for yxt nlp processing',
    long_description=readme(),
    keywords='yunxuetang nlp',
    author='wanglijun',
    author_email='juns1984@qq.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Text Processing :: Linguistic',
    ],
    url='https://github.com/junix/yxt_nlp_toolkit.git',
    packages=[
        'yxt_nlp_toolkit',
        'yxt_nlp_toolkit.common',
        'yxt_nlp_toolkit.utils',
        'yxt_nlp_toolkit.utils.jieba_dict',
        'yxt_nlp_toolkit.embedding',
    ],
    install_requires=[
        # 'gensim==3.4.0',
        'jieba',
    ],
    include_package_data=True,
    zip_safe=False
)
